#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# BraggLaw.py
#
# $Id:    $
# $URL: $
#
# Part of the "pydiffract" package


import sys
import math
# import LatticeSym		# note, if you don't need LatticeSym, then just commnent out this line

__version__ = "$Revision: $"
__author__  = "Jon Z. Tischler, <tischler@aps.anl.gov>" +\
              "Argonne National Laboratory"
__date__    = "$Date: $"
__id__      = "$Id: $"

hc_keVnm = 1.2398419739				# h*c (keV-nm)
nan = float('nan')


class BraggLaw(object):
	""" 
	Class that uses Bragg's Law to update a calculation of Bragg's Law when a given parameter is changed.
	e.g. if energy is changed, then it updates theta.
	It accepts energy in place of wavelength, and 2theta in place of theta

	wavelength		wavelength (nm)
	d				d-spacing (nm)
	theta			Bragg angle (degree)
	"""

	def __init__(self, keV=10, d=0.3):
		"""
		Initialize the BraggLaw instance.
		keV		energy
		d		d-spacing (nm)
		"""
		global hc_keVnm

		try:	keV = float(keV)
		except:	keV = 10.0
		try:	d = float(d)
		except:	d = 0.3

		self.energy = keV
		self.d = d
		self.wavelength = hc_keVnm / keV

		try:	self.theta = math.asin( self.wavelength / (2 * d) ) * 180/(math.pi)
		except: self.theta = nan
		self.tth = 2 * (self.theta)

		self.newBit = 0		# 'wavelength'
		self.oldBit = 1		# 'd'
		return None


	def __str__(self):
		""" Return string value for BraggLaw. """
		out = '%.7g keV (%.7g nm),   d = %.5g nm,   theta = %.6g° (2theta = %.6g°)' % (self.energy, self.wavelength, self.d, self.theta, self.tth)
		return out


	def __repr__(self):
		""" Return representation value for atomXtal. """
		out = self.__str__()
		out += ',   new = "%s",   old = "%s"' % (self.__nameOfBit(self.newBit),self.__nameOfBit(self.oldBit))
		return out


	def updateFromOne(self, name, val):
		"""
		From name, and self.oldBit, & self.newBit, compute new value.
		Note, self.newBit always takes on value of last one changed (i.e. self.newBit == ibit)
		bits:   0 == (wavelength or energy),   1 == d-spacing,   2 == (theta or 2theata)
		"""
		global hc_keVnm

		try:	val = float(val)
		except:	return None

		try:
			ibit = int(name)							# name can also be a bit value of 0, 1, 2
		except:
			name = name.lower()
			ibit = -1

		if ibit == 0:
			self.wavelength = val
		elif ibit == 1:
			self.d = val
		elif ibit == 2:
			self.theta = val
		elif name in set(['lambda','nm','wavelength']):	# aliases for wavelength
			self.wavelength = val
			ibit = 0
		elif name in ['energy','keV']:					# pretend that energy was really wavelength
			self.wavelength = hc_keVnm / val
			ibit = 0

		elif name.startswith('d'):						# aliases: d-spacing is anything starting with 'd'
			self.d = val
			ibit = 1

		elif name.startswith('theta'):
			self.theta = val
			ibit = 2
		elif name in ['tth','2theta','2-theta','2th']:	# pretend that 2theta was really theta
			self.theta = val / 2
			ibit = 2

		if ibit < 0: return None

		if ibit != self.newBit:							# a different ibit, push down stack
			self.oldBit = self.newBit
			self.newBit = ibit

		# update the value that needs to be changed
		if   (ibit + self.oldBit) == 1:	self.__calc_theta()			# ibit = 0 + 1
		elif (ibit + self.oldBit) == 2:	self.__calc_d()				# ibit = 0 + 2
		elif (ibit + self.oldBit) == 3:	self.__calc_wavelength()	# ibit = 1 + 2

		self.energy = hc_keVnm / self.wavelength
		self.tth = 2 * self.theta
		return None


	def __calc_theta(self):
		""" wavelength & d --> theta """
		try:	self.theta = math.asin( self.wavelength / (2 * self.d) ) * 180/(math.pi)
		except:	self.theta = nan
		return self.theta


	def __calc_d(self):
		""" wavelength & theta --> d """
		self.d = self.wavelength / ( 2 * math.sin(self.theta * (math.pi)/180) )
		return self.d


	def __calc_wavelength(self):
		""" theta & d --> wavelength """
		self.wavelength = 2 * self.d * math.sin(self.theta * (math.pi)/180)
		return self.wavelength


	def __nameOfBit(self,ibit):
		if   ibit == 0: return 'wavelength'
		elif ibit == 1: return 'd'
		elif ibit == 2: return 'theta'
		return ''




if __name__ == '__main__':
	"""
	Main function for BraggLaw.py.
	Test cases for BraggLaw class to verify correct behavior.
	"""
	hasLatticeSym = 'LatticeSym' in locals().keys()
	maxBit = 4
	if hasLatticeSym: 	maxBit = 16

	try:	testGroup = int(sys.argv[1])
	except:	raise ValueError('./BraggLaw testGroup,  testGroup must be a bit mask, a combination of 1-'+str(maxBit)+', use -1 for all')
	print 'showing testGroup(s) == %r (= %r)' % (testGroup,bin(testGroup))

	iflag = 1
	if testGroup & iflag:
		print '\n---testGroup',iflag
		try:	bl = BraggLaw()
		except:	raise ValueError('./BraggLaw,  cannot initialize BraggLaw')
		print repr(bl)

		bl.updateFromOne('energy', 12)
		print repr(bl)

		bl.updateFromOne('d', 0.3)
		print repr(bl)

		bl.updateFromOne('d', 0.30001)
		print repr(bl)

		bl.updateFromOne('2theta', 23.0)
		print repr(bl)

	iflag *= 2
	if testGroup & iflag:
		print '\n---testGroup',iflag
		try:	bl = BraggLaw(keV=10.36477)
		except:	raise ValueError('./BraggLaw,  cannot initialize BraggLaw')
		print repr(bl)

	iflag *= 2
	if testGroup & iflag:
		print '\n---testGroup',iflag
		try:	bl = BraggLaw(keV=10.36477)
		except:	raise ValueError('./BraggLaw,  cannot initialize BraggLaw')
		print repr(bl)


	if hasLatticeSym:
		iflag *= 2
		if testGroup & iflag:
			print '\n---testGroup',iflag
			print 'Using Silicon:'
			atomSi = LatticeSym.atomXtal('Si1',0,0,0,WyckoffSymbol='a',DebyeT=645)
			silicon = LatticeSym.Lattice(227, 0.54310206,0,0, 0,0,0, desc='Silicon',Temperature=22.5,alphaT=2.56e-6,atoms=(atomSi,))

			try:	bl = BraggLaw(keV=10, d=silicon.dSpacing(1,1,1))
			except:	raise ValueError('./BraggLaw,  cannot initialize BraggLaw')
			print repr(bl)

		iflag *= 2
		if testGroup & iflag:
			print '\n---testGroup',iflag
			print 'Using Diamond:'
			atomC = LatticeSym.atomXtal('C1',0,0,0,WyckoffSymbol='a')
			diamond = LatticeSym.Lattice(227, 0.3567,0,0, 0,0,0, desc='Diamond', atoms=(atomC,))

			try:	bl = BraggLaw(keV=10, d=diamond.dSpacing(1,0,0))
			except:	raise ValueError('./BraggLaw,  cannot initialize BraggLaw')

			print repr(bl)
			bl.updateFromOne('d', diamond.dSpacing(0,0,4))
			print repr(bl)

