#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""	GUI for Bragg Law Panel, APS
	written by Jon Tischler (XSD, Argonne National Laboratory)
"""


import os
import sys
import datetime
import time
import wx
# import wx.html

import JZTutil
import BraggLaw
import LatticeSym

nan = float('nan')
VERSION_STR = '1.0.1'
Temperature = 300.		# temperature (K)
GREEK_SMALL_THETA = u'\u03B8'		# unicode for theta
GREEK_SMALL_LAMBDA = u'\u03BB'		# unicode for lambda
DEGREE_SIGN = u'\u00B0'				# unicode for degree sign
SUPERSCRIPT_TWO = u'\u00B2'			# super script 2, used to show |F|^2


class MyApp(wx.App):
	""" Application class. """

	def OnInit(self):
		if sys.platform.startswith('win32'):
			mainFrameSize = (480, 452)  # needs to be larger on Windows, with bin path (540, 480)
		elif sys.platform.startswith('linux'):
			mainFrameSize = (450, 407)  # size for Linux on hpcs34, with bin path (480, 370)
		elif sys.platform.startswith('darwin'):
			mainFrameSize = (240, 262)  # was (480, 370)
		else:
			mainFrameSize = (420, 342)  # size for Mac, with bin path (480, 370)

		self.frame = MainFrame("Bragg's Law", (50,60), mainFrameSize)
		self.frame.Show()
		self.SetTopWindow(self.frame)
		return True



class MainFrame(wx.Frame):
	""" Main Frame class that it contains top level items. """

	def __init__(self, title, pos, size):
		wx.Frame.__init__(self, None, -1, title, pos, size)

		backgroundColor = self.GetBackgroundColour()
		self.backgroundColor = (backgroundColor[0], backgroundColor[1], backgroundColor[2], 0)

		self.CreateStatusBar()
		self.SetStatusText('')

		panel = wx.Panel(self)
		mainSizer = wx.BoxSizer(wx.VERTICAL)
		panel.SetSizer(mainSizer)

		try:	self.bl = BraggLaw.BraggLaw(keV=10, d=0.3)
		except:	raise ValueError('./BraggLaw,  cannot initialize BraggLaw')

		self.setUpMenus()
		fontSize = 18
		numberWidth = 120			# was 170

		# wx.Font(pointSize, family, style, weight, underline, face, encoding) 
		regularFont = wx.Font(fontSize, wx.SWISS, wx.NORMAL, wx.NORMAL)
		hklFont = wx.Font(fontSize, wx.SCRIPT, wx.FONTSTYLE_ITALIC, wx.NORMAL)
		smallFont = wx.Font(fontSize-6, wx.ROMAN, wx.NORMAL, wx.NORMAL)

		self.xtl = None
		self.h = 1
		self.k = self.l = 0
		hkl = [self.h,self.k,self.l]
		hklStr = JZTutil.hkl2str(hkl)
		self.Ftxt = ' '

		# start of the general controls
		self.energyLabel = wx.StaticText(panel, -1, 'E (kev)  ')
		self.energyLabel.SetFont(regularFont)
		self.energyStr = wx.TextCtrl(panel, -1, str(self.bl.energy), size=(numberWidth,-1), style=wx.TE_PROCESS_ENTER) 
		self.energyStr.SetFont(regularFont)
		energySizer = wx.BoxSizer(wx.HORIZONTAL)
		energySizer.Add(self.energyLabel,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
		energySizer.Add(self.energyStr,1,wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)
		mainSizer.Add(energySizer,0,wx.ALIGN_LEFT|wx.EXPAND|wx.ALL,5)
		self.wavelengthLabel = wx.StaticText(panel, -1, GREEK_SMALL_LAMBDA + u' (nm)  ')
		self.wavelengthLabel.SetFont(regularFont)
		self.wavelengthStr = wx.TextCtrl(panel, -1, str(self.bl.wavelength), size=(numberWidth,-1), style=wx.TE_PROCESS_ENTER)
		self.wavelengthStr.SetFont(regularFont)
		wavelengthSizer = wx.BoxSizer(wx.HORIZONTAL)
		wavelengthSizer.Add(self.wavelengthLabel,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
		wavelengthSizer.Add(self.wavelengthStr,1,wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)
		mainSizer.Add(wavelengthSizer,0,wx.ALIGN_LEFT|wx.EXPAND|wx.ALL,5)

		self.thetaLabel = wx.StaticText(panel, -1, u'      ' + GREEK_SMALL_THETA + DEGREE_SIGN + u'  ')
		self.thetaLabel.SetFont(regularFont)
		self.thetaStr = wx.TextCtrl(panel, -1, str(self.bl.theta), size=(numberWidth,-1), style=wx.TE_PROCESS_ENTER)
		self.thetaStr.SetFont(regularFont)
		thetaSizer = wx.BoxSizer(wx.HORIZONTAL)
		thetaSizer.Add(self.thetaLabel,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
		thetaSizer.Add(self.thetaStr,1,wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)
		mainSizer.Add(thetaSizer,0,wx.ALIGN_LEFT|wx.EXPAND|wx.ALL,5)

		self.tthLabel = wx.StaticText(panel, -1, u'    2' + GREEK_SMALL_THETA + DEGREE_SIGN + u'  ')
		self.tthLabel.SetFont(regularFont)
		self.tthStr = wx.TextCtrl(panel, -1, str(self.bl.tth), size=(numberWidth,-1), style=wx.TE_PROCESS_ENTER)
		self.tthStr.SetFont(regularFont)
		tthSizer = wx.BoxSizer(wx.HORIZONTAL)
		tthSizer.Add(self.tthLabel,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
		tthSizer.Add(self.tthStr,1,wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)
		mainSizer.Add(tthSizer,0,wx.ALIGN_LEFT|wx.EXPAND|wx.ALL,5)

		self.dLabel = wx.StaticText(panel, -1, 'd (nm)  ')
		self.dLabel.SetFont(regularFont)
		self.dStr = wx.TextCtrl(panel, -1, str(self.bl.d), size=(numberWidth,-1), style=wx.TE_PROCESS_ENTER)
		self.dStr.SetFont(regularFont)
		dSizer = wx.BoxSizer(wx.HORIZONTAL)
		dSizer.Add(self.dLabel,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
		dSizer.Add(self.dStr,1,wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)
		mainSizer.Add(dSizer,0,wx.ALIGN_LEFT|wx.EXPAND|wx.ALL,5)

		self.dButton = wx.Button(panel, label='Get d')
		self.dButton.Enable()
		self.hklLabel = wx.StaticText(panel, -1, 'hkl')
		self.hklLabel.SetFont(hklFont)
		self.hklStr = wx.TextCtrl(panel, -1, hklStr, size=(130,-1), style=wx.TE_PROCESS_ENTER)
		self.hklStr.SetFont(regularFont)
		getdSizer = wx.BoxSizer(wx.HORIZONTAL)
		getdSizer.Add(self.dButton,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL|wx.RIGHT,5)
		getdSizer.Add(self.hklLabel,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
		getdSizer.Add(self.hklStr,0,wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL|wx.EXPAND)
		mainSizer.Add(getdSizer,0,wx.EXPAND|wx.ALIGN_LEFT|wx.ALL,5)

 		self.xtalDesc = wx.StaticText(panel, -1, ' ')
 		self.xtalDesc.SetFont(smallFont)
 		mainSizer.Add(self.xtalDesc,0,wx.EXPAND|wx.ALIGN_LEFT|wx.ALL,1)
 
 		self.FhklStr = wx.StaticText(panel, -1, ' ')
 		self.FhklStr.SetFont(smallFont)
		mainSizer.Add(self.FhklStr,0,wx.EXPAND|wx.ALIGN_LEFT|wx.ALL,2)

		self.Bind(wx.EVT_TEXT_ENTER, self.OnEnergyChange,self.energyStr)
		self.Bind(wx.EVT_TEXT_ENTER, self.OnWavelengthChange,self.wavelengthStr)
		self.Bind(wx.EVT_TEXT_ENTER, self.OnThetaChange,self.thetaStr)
		self.Bind(wx.EVT_TEXT_ENTER, self.On2ThetaChange,self.tthStr)
		self.Bind(wx.EVT_TEXT_ENTER, self.OndChange,self.dStr)
		self.Bind(wx.EVT_BUTTON, self.OnGetd,self.dButton)
		self.Bind(wx.EVT_TEXT_ENTER, self.OnHKLchange,self.hklStr)

		self.updatePanel()

	def OnEnergyChange(self, event):
		try:	energy = float(self.energyStr.GetValue())
		except:	energy = nan
		self.bl.updateFromOne('energy',energy)
		self.updatePanel()
		txt = 'energy set to %g (keV)' % (energy,)
		self.SetStatusText(txt)
		print 'OnEnergyChange:  ' + txt

	def OnWavelengthChange(self, event):
		try:	wavelength = float(self.wavelengthStr.GetValue())
		except:	wavelength = nan
		self.bl.updateFromOne('wavelength',wavelength)
		self.updatePanel()
		txt = 'wavelength set to %g (keV)' % (wavelength,)
		self.SetStatusText(txt)
		print 'OnWavelengthChange:  ' + txt

	def OnThetaChange(self, event):
		try:	theta = float(self.thetaStr.GetValue())
		except:	theta = nan
		self.bl.updateFromOne('theta',theta)
		self.updatePanel()
		self.SetStatusText((u'theta set to %g'+DEGREE_SIGN) % (theta,))
		print 'OnThetaChange:  theta --> %g°' % (theta,)

	def On2ThetaChange(self, event):
		try:	tth = float(self.tthStr.GetValue())
		except:	tth = nan
		self.bl.updateFromOne('2theta',tth)
		self.updatePanel()
		self.SetStatusText((u'2-theta set to %g'+DEGREE_SIGN) % (tth,))
		print 'On2ThetaChange:  2theta --> %g°' % (tth,)

	def OndChange(self, event):
		try:	d = float(self.dStr.GetValue())
		except:	d = nan
		self.bl.updateFromOne('d',d)
		self.hklStr.SetValue('')		# d-spacing set by hand, so hkl is vague
		self.Ftxt = ' '
		self.updatePanel()
		txt = 'd set to %g (nm)' % (d,)
		self.h = self.k = self.l = nan
		self.SetStatusText(txt)
		print 'OndChange:  ' + txt

	def OnGetd(self, event):
		self.OnOpenXTL()
		try: 	txt = 'Get d Button pushed, using:  ' + self.xtl.desc
		except:	txt = 'Get d Button pushed, get the d-spacing  NO xtl found'
		self.SetStatusText(txt)
		try:
			d = float(self.xtl.dSpacing(self.h,self.k,self.l))
			self.setFstr()
		except:
			d = nan
			self.Ftxt = ' '

		self.bl.updateFromOne(1,d)
		print self.bl
		self.updatePanel()
		txt = 'd set to %g (nm)' % (d,)
		self.SetStatusText(txt)
		print 'OnGetd:  ' + txt


	def OnHKLchange(self, event):
		""" The hkl was changed. """
		try:	hklStr = self.hklStr.GetValue()
		except:	hklStr = ''
		try:
			hkl = JZTutil.str2hkl(hklStr)
			self.h = hkl[0]
			self.k = hkl[1]
			self.l = hkl[2]
			d = float( self.xtl.dSpacing(self.h,self.k,self.l) )
			self.setFstr()
		except:
			d = nan
			self.Ftxt = ' '

		self.bl.updateFromOne('d',d)
		self.updatePanel()
		txt = 'd(%s) set to %g (nm)' % (hklStr,d)
		self.SetStatusText(txt)
		print 'OnHKLchange:  ' + txt


	def setFstr(self):
		try:
			Fhkl =  self.xtl.Fstruct(self.h,self.k,self.l,keV=self.bl.energy,T_K=Temperature)
			Fabs = abs(Fhkl)
			if Fabs > 1e-2:
				self.Ftxt = 'Fhkl = %s + i*%s,   |F|%s = %s' % (self.rfmt(Fhkl.real,'%.2f'),self.rfmt(Fhkl.imag,'%.2f'),SUPERSCRIPT_TWO,self.rfmt(Fabs**2,'%.2f'))
			elif Fabs == Fabs:
				self.Ftxt = 'Fhkl = 0'
			else:
				sel.Ftxt = ' '
		except:
			self.Ftxt = ' '


	def updatePanel(self):
		""" update all of the value in the panel """
		self.energyStr.SetValue(self.rfmt(self.bl.energy,'%.4f'))
		self.wavelengthStr.SetValue(self.rfmt(self.bl.wavelength,'%.6f'))
		self.thetaStr.SetValue(self.rfmt(self.bl.theta,'%.4f'))
		self.tthStr.SetValue(self.rfmt(self.bl.tth,'%.4f'))
		self.dStr.SetValue(self.rfmt(self.bl.d,'%.6f'))
		if self.bl.newBit + self.bl.oldBit == 2: self.Ftxt=' '	# set by E & angle
		self.FhklStr.SetLabel(self.Ftxt)
		self.setLabelColors()


	def rfmt(self, val, fmt):
		if val.is_integer():
			sval = str(int(val))
		else:
			sval = fmt % (val,)
			sval = sval.rstrip('0')
			sval = sval.rstrip('.')
		return sval

	def setLabelColors(self):
		white = (255,255,255)	# white
		black = (0,0,0)			# black
		darkBlue = (0,50,204)	# dark blue
		lightBlue = (0,135,204)	# light blue

		self.energyLabel.SetForegroundColour(black)		# start with all black, then set newBit and oldBit
		self.wavelengthLabel.SetForegroundColour(black)
		self.thetaLabel.SetForegroundColour(black)
		self.tthLabel.SetForegroundColour(black)
		self.dLabel.SetForegroundColour(black)

		# 0 == wavelength & energy,   1 == d-spacing,   2 == theta & 2theta
		if self.bl.newBit == 0:
			self.energyLabel.SetForegroundColour(darkBlue)
			self.wavelengthLabel.SetForegroundColour(darkBlue)
		elif self.bl.newBit == 2:
			self.thetaLabel.SetForegroundColour(darkBlue)
			self.tthLabel.SetForegroundColour(darkBlue)
		else:
			self.dLabel.SetForegroundColour(darkBlue)

		if self.bl.oldBit == 0:
			self.energyLabel.SetForegroundColour(lightBlue)
			self.wavelengthLabel.SetForegroundColour(lightBlue)
		elif self.bl.oldBit == 2:
			self.thetaLabel.SetForegroundColour(lightBlue)
			self.tthLabel.SetForegroundColour(lightBlue)
		else:
			self.dLabel.SetForegroundColour(lightBlue)

		if self.xtl:
			self.hklStr.Enable()
			self.hklStr.SetBackgroundColour(white)
			self.hklLabel.SetForegroundColour(black)
			self.hklStr.SetForegroundColour(black)
			if (self.bl.newBit + self.bl.oldBit) == 2:
				self.hklStr.SetValue('')	# d-spacing set by E & angle, so hkl is vague
		else:
			self.hklStr.Disable()
			self.hklStr.SetBackgroundColour(self.backgroundColor)
			self.hklStr.SetForegroundColour(self.backgroundColor)
			self.hklLabel.SetForegroundColour(self.backgroundColor)


	def setUpMenus(self):
		# set up the File menu in the menubar
		helpMenu = wx.Menu()
		helpMenu.Append(wx.ID_ABOUT, "About BraggPanel", "Display information about the program")
		helpMenu.Append(wx.ID_HELP, "Help")
		# for a list of standard items see:  http://www.wxpython.org/docs/api/wx.Button-class.html
		wx.EVT_MENU(self, wx.ID_ABOUT, self.OnAbout)
		wx.EVT_MENU(self, wx.ID_HELP, self.OnHelp)
		wx.EVT_MENU(self, wx.ID_EXIT, self.OnQuit)

		# define menubar, and put help menu into the menubar
		menuBar = wx.MenuBar()
		menuBar.Append(helpMenu, "&Help")
		self.SetMenuBar(menuBar)


	def OnOpenXTL(self):
		""" Open an xtl file. Start at the materials/ folder"""
		dirname = os.path.split(os.path.realpath(__file__))[0]
		dirname = os.path.join(dirname,'materials')
		wildcard = 'crystal files (*.xml)|*.xml|CIF files (*.cif)|*.cif|any file (*.*)|*.*'
#		print 'file open dialog with dirname = "%s"' % (dirname,)
		dlg = wx.FileDialog(self, "Choose a file", dirname, "", wildcard, wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
		if dlg.ShowModal() == wx.ID_OK:
			filename = dlg.GetFilename()
			dirname = dlg.GetDirectory()
			print 'xml from:  filename=',filename,'   dirname =',dirname
			dlg.Destroy()
			self.xtl = LatticeSym.Lattice(0, 0,0,0,0,0,0,file=os.path.join(dirname, filename))
			self.xtalDesc.SetLabel(self.xtl.desc)


	def OnAbout(self, event):
		self.SetStatusText('called About')
		print 'at top of OnAbout'
		txt = "This is a caluculator for Bragg's Law."
		txt += '\rLast modified: ' + time.strftime("%a, %b %d %Y, %I:%M %p", time.localtime(os.path.getmtime(__file__)))+'.'
		txt += '\r\twritten by Jon Tischler'
		txt += '\r\t(XSD/APS, Argonne National Laboratory)'
		wx.MessageBox(txt, 'BraggLawPanel (v'+str(VERSION_STR)+')', wx.OK | wx.ICON_INFORMATION, self)

	def OnQuit(self, event):
		print 'Quitting'
		self.CloseEverything()

	def CloseEverything(self):
		""" Closes all windows, leave nothing open,  and quit all. """
		print 'Close all windows & Quit'
		self.Close()


 	def OnHelp(self, event):
		print 'Help Goes Here'
 		self.SetStatusText('called Help, not yet available')
# 		self.HelpFrame = HelpFrame()
# 		self.HelpFrame.Show()






# class HelpFrame(wx.Frame):
# 	""" Frame used to show html help. """
# 
# 	def __init__(self):
# 		wx.Frame.__init__(self, None, -1, "Help", size=(800, 500))
# 		panel = wx.Panel(self, -1)
# 		vbox = wx.BoxSizer(wx.VERTICAL)
# 		htmlwin = wx.html.HtmlWindow(panel, -1, style=wx.NO_BORDER)
# 		htmlwin.SetBackgroundColour(wx.RED)
# 		htmlwin.SetBorders(10)
# 		htmlwin.SetStandardFonts()
# 
# 		dir= os.path.dirname(os.path.realpath(__file__))	# directory to 'ORNLclient.py'
# 		htmlHelpFile = os.path.join(dir,'Server Help.html')
# 		if not os.path.exists(htmlHelpFile): htmlHelpFile = 'Server Help.html'
# 		htmlwin.LoadPage(htmlHelpFile)
# 
# 		vbox.Add((-1, -1), 0)
# 		vbox.Add(htmlwin, 1, wx.EXPAND | wx.ALL, 5)
# 		panel.SetSizer(vbox)
# 		self.Centre()
# 		self.Show(True)
# 
# 
# 	def OnClose(self, event):
# #		self.Destroy()
# 		self.Close()









if __name__ == '__main__':
	# log file for this client
	if sys.platform == 'darwin':
		# for Mac OSX, put logs into offical Logs location
		logDir = os.path.expanduser('~/Library/Logs/gov.anl.aps.SSM/')
		if not os.path.exists(logDir):
			try:
				os.makedirs(logDir)
			except:
				print 'ERROR -- Unable to create directory for Log file = "%r",' % logDir
				sys.exit(1)
	else:
		logDir = os.path.expanduser('~')
	clientLogName = os.path.join(logDir,'BraggLawPanel.log')


	startStr = 'BarggLaw Panel:   '+datetime.datetime.now().strftime('%a, %B %d, %Y %I:%M:%S%p')
	try:	clientLog = open(clientLogName,'a',1)	# open as an existing file for appending & line buffered
	except:
		print 'ERROR, cannot open client log file "'+clientLogName+'"\n'
		sys.exit(1)
	print startStr
	print '	redirecting stdout & stderr to log file', clientLogName


	saveout = sys.stdout				# save in case I want to un-redirect stdout
	sys.stdout = clientLog				# re-direct stdout to clientLog file
	sys.stderr = clientLog				# re-direct stderr to clientLog file
	print '\n\n******************************************************************************'
	print '******************************************************************************'
	print '******************************************************************************'
	print startStr
	print 'python version  =  '+sys.version+'\n\n'

	app = MyApp(False)
	app.MainLoop()

	if type(clientLog) is file:
		doneStr = '\nBraggLawPanel Ended:   '+datetime.datetime.now().strftime('%a, %B %d, %Y %I:%M:%S%p')+'\n'
		print doneStr
		clientLog.close()
		sys.stdout = saveout				# re-direct stdout to original stream
		sys.stderr = saveout				# re-direct stderr to original stream
		print doneStr						# prints to console

